/*
compile: gcc -lgslcblas -lgsl test_curve_ci.c
run:     echo "1 1 1 1 1" | test_curve_ci
*/

#include <math.h>
#include <stdio.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_odeiv.h>

#define M 4 /* number of kinetic coefficients */
#define N 3 /* number of chemical components = order of the diff. eqn. */

typedef struct {
    double k[M];
} kin_par_type;

int func( double t, const double y[], double dydt[], void *params )
{
    kin_par_type *P = (kin_par_type*) params;
    double *k = P->k;

    dydt[0] = -y[0]*(k[0]+k[1]*y[1]);
    dydt[2] =  y[1]*(k[2]+k[3]*y[2]);
    dydt[1] =  -dydt[0] -dydt[2];
    return GSL_SUCCESS;
}

// fit: 0   x0            x0.5          x0            x0             -1.128308    x1            x0            x0             0.0224332     0.408077      0.01546256    1.436748      1.637468      0.1197546     0.6570706    


int main( int argc, char **argv )
{
    int i, m, ret;
    double t, y[N], outfrac[N], k[M], t0, t1, bg;
    kin_par_type P;

    double duration = 200.0;
    double h = 1e-6;

    gsl_odeiv_system sys = { func, NULL, N, &P };
    const gsl_odeiv_step_type *T = gsl_odeiv_step_rkf45; // 8pd;

    gsl_odeiv_step *s  = gsl_odeiv_step_alloc (T, N);
    gsl_odeiv_control *c = gsl_odeiv_control_y_new (1e-12, 0.0);
    gsl_odeiv_evolve *e = gsl_odeiv_evolve_alloc (N);

    // Start values:
    y[0] = 1;
    for( i=1; i<N; ++i )
        y[i] = 0;

    // Read parameters from stdin:
    // fprintf( stderr, "# test_curve_ci\n" );
    scanf( "%i", &m );
    if ( m!= M+1 ) {
        fprintf( stderr,
                 "inconsistent number of parameters: %i offered, %i needed\n",
                 m, M+1 );
        return -1;
    }
    scanf( "%lg", &t0 );
    // fprintf( stderr, "# t0 %g\n", t0 );
    for ( m=0; m<M; ++m ) {
        double val;
        scanf( "%lg", &val );
        P.k[m] = val>0 ? val : 0;
        // fprintf( stderr, "# k%i %g\n", m, P.k[m] );
    }
    // fprintf( stderr, "# start evaluation\n" );

    t = t0;
    t1 = t0 + duration;
    int nout=0;
    while (t < t1) {
        ret = gsl_odeiv_evolve_apply( e, c, s, &sys, &t, t1, &h, y);
        if ( ret != GSL_SUCCESS ){
            fprintf( stderr, "gsl errno %i\n", ret );
            break;
        }
        printf ("%9.5f %12.5e\n", t, y[1] );
        ++nout;
    }
    // fprintf( stderr, "# wrote %i data pairs\n", nout );

    gsl_odeiv_evolve_free (e);
    gsl_odeiv_control_free (c);
    gsl_odeiv_step_free (s);
    return 0;
}
